
%if 0

Resident code of SEEKEXT
 2020 by C. Masloch

Usage of the works is permitted provided that this
instrument is retained with the works, so that any entity
that uses the works is notified of this instrument.

DISCLAIMER: THE WORKS ARE WITHOUT WARRANTY.

%endif

%if 0

Supported Int21 functions:

Seek file handle
INP:	ah = 42h
	al = seek origin:
	 SEEK_SET 0
	 SEEK_CUR 1
	 SEEK_EOF 2
	cx:dx = offset
OUT:	NC if success,
	 dx:ax = new seek position (absolute)
	CY if error,
	 ax = error code
Note:	Hooked by SEEKEXT to handle old-style seeking
	 in terms of new-style seeking.

Long seek file handle
INP:	ax = 7142h
	cl = seek origin
	ds:dx -> offset qword structure
OUT:	NC if success,
	 ds:dx -> structure updated to new seek position
	CY if error,
	 ax = error code (7100h unsupported)
	CF unchanged if 71h interface not supported,
	 ax = 7100h
Note:	Hooked by SEEKEXT to pass new-style seeking
	 to redirectors (dosemu2 extension service
	 2F.1142).


Supported Int2D functions:

AMIS - Installation check
INP:	al = 00h
OUT:	al = 0FFh
	cx = Private version number (currently 0100h)
	dx:di-> signature: "ecm     ","SEEKEXT "

AMIS - Get private entry point - NOP: no private entry point
INP:	al = 01h
OUT:	al = 00h

AMIS - Uninstall - NOP: no resident uninstaller or NOP: can't uninstall
INP:	al = 02h
OUT:	If installed from command line,
	 al = 03h
	 bx = memory block of resident TSR (cs)
	(If internal (pre-installed),
	 al = 01h
	If with resident uninstaller,
	 al = 01h if unsuccessful
	 al = 0FFh if successful)

AMIS - Request pop-up - NOP: no pop-up
INP:	al = 03h
OUT:	al = 00h

AMIS - Determine chained interrupts
INP:	al = 04h
OUT:	al = 04h
	dx:bx-> interrupt hook list (Int2D always.)

AMIS - Get hotkeys - NOP: no hotkeys
INP:	al = 05h
OUT:	al = 00h

AMIS - Get device driver information - NOP: no device
INP:	al = 06h
OUT:	al = 00h

AMIS - Reserved for AMIS
INP:	al = 07h..0Fh
OUT:	al = 00h

TSR - Get compatible version
INP:	al = 10h
OUT:	al = 0FFh
	cx = compatible version number (currently 0100h)

TSR - Get ctrl0
INP:	al = 20h
OUT:	al = size of returned control data (not 0)
	dx:bx-> control data
	Unsupported bits switched off

TSR - Reserved for TSR
INP:	al = 11h..1Fh, 21h..FFh
OUT:	al = 00h

%endif

%if TSR_VERP != ver(1,00)		; if defined one doesn't match current
 %error "TSR version doesn't match"
%endif

handler:
				; AMIS stuff
amissig:
.ven:	fill 8,32,db "ecm"		; vendor
.prod:	fill 8,32,db "SEEKEXT"		; product
.desc:	asciz "DOS seek service extension"	; description
%if $ - .desc > 64
 %error AMIS description too long
%endif

amisintr:
.i21:	db 21h
	dw i21
.i2D:	db 2Dh
	dw i2D

				; TSR data
ctrl0:	db 1100_0000b			; control byte
					; bit 6: state 1
					;     7: state 2

i2D.uninstall:
	inc ax				; (= 03h) safe to remove but no resident uninstaller
	mov bx, cs			; = segment
i2D.hwreset equ $-1		; (second byte of mov bx, cs is same as the retf opcode)
	iret

iispentry i2D, 0, i2D
	cmp ah, 0
amisnum equ $-1				; AMIS multiplex number (data for cmp opcode)
	je .handle			; our multiplex number -->
	jmp far [cs:.next]		; else go to next handler -->

.handle:
	test al, al
	jz .installationcheck		; installation check -->
	cmp al, 02h
	je .uninstall			; uninstallation -->
	cmp al, 04h
	je .determineinterrupts		; determine hooked interrupts -->
	cmp al, 10h
	je .getver			; get compatible version -->
	cmp al, 20h
	je .getctrl0			; get ctrl0 -->
				; all other functions are reserved or not supported by TSR
.nop:
	mov al, 0			; show not implemented
	iret

.installationcheck:
	dec al				; (= FFh) show we're here
	mov cx, TSR_VERP		; = version
	xor di, di			; dx:di -> AMIS signature strings of this program
.iret_dx_cs:
	mov dx, cs
.iret:
	iret

.determineinterrupts:			; al = 04h, always returns list
	mov bx, amisintr		; dx:bx -> hooked interrupts list
	jmp short .iret_dx_cs

.getver:
	mov al, 0FFh			; show call supported
	mov cx, TSR_VERC		; = compatible version
	iret

.getctrl0:
	mov bx, ctrl0			; dx:bx -> ctrl0
	mov al, 1			; size of ctrl0: 1 byte
	and byte [cs:bx], _state_short | _state_long
	jmp short .iret_dx_cs


iispentry i21, 0, i2D
	cmp ax, 7142h
	je .long_check
	cmp ah, 42h
	jne .chain
.short_check:
	test byte [cs:ctrl0], _state_short
	jnz .short
.chain:
	jmp far [cs:.next]


.long_check:
	test byte [cs:ctrl0], _state_long
	jz .chain
.long:
	push es
	push di
	push bx
	mov ax, 1220h
	stc
	int 2Fh			; get PHT entry from handle
	jc .pop_iret_CF
	mov bl, byte [es:di]
	mov bh, 0
	cmp bl, -1
	mov ax, 6		; invalid handle error code
	stc
	je .pop_iret_CF
	mov ax, 1216h
	; stc
	int 2Fh			; get SFT entry from PHT handle
	jc .pop_iret_CF
	mov ax, 7142h		; restore function number
	test byte [es:di + 6], 80h
	jz .pop_chain
	mov ah, 11h		; ax = 1142h
	stc
	int 2Fh
.pop_iret_CF:
	pop bx
	pop di
	pop es
.iret_CF:
	push bp
	mov bp, sp
	rcr byte [bp + 6], 1	; flip
	rol byte [bp + 6], 1	; flop
	pop bp
	iret

.pop_chain:
	pop bx
	pop di
	pop es
	jmp .chain


.short:
	lframe int
	lenter
	lvar word,	frame_ds
	 push ds
	lvar word,	frame_cx
	 push cx
	lvar word,	frame_dx
	 push dx
	lequ ?frame_dx,	frame_cxdx
	lvar word,	function
	 push ax
	test al, al		; seek set ?
	mov ax, 0
	jz @F			; yes, zero-extend -->
	test ch, ch		; sign bit set ?
	jns @F			; no, zero-extend -->
	dec ax			; else, sign-extend
@@:
	lvar qword,	seekstructure
	 push ax
	 push ax		; sign or zero extension
	 push cx
	 push dx		; low 32 bits
	push ss
	pop ds
	mov dx, sp		; -> seekstructure
	mov cl, byte [bp + ?function]
	mov ax, 7142h
	stc
	int 21h
	jc .short_chain

	mov ax, -1
	mov dx, ax		; report -1 if out of range

	mov cx, word [bp + ?seekstructure + 6]
	jcxz @F			; high bits all zeros ? -->
	cmp cx, ax		; high bits all ones ?
	jne .short_return	; no, report -1 now -->
@@:
	cmp word [bp + ?seekstructure + 4], cx
				; do the next high bits match ?
	jne .short_return	; no, report -1 now -->
	mov dx, word [bp + ?seekstructure + 2]
	mov ax, word [bp + ?seekstructure]
				; dx:ax = reported short seek

.short_return:
	add sp, 8		; discard structure
	clc
	pop cx			; discard ax
	pop cx			; discard dx
	pop cx			; restore cx
	pop ds			; restore ds
	lleave code
	jmp .iret_CF		; return NC

.short_chain:
	add sp, 8		; discard structure
	pop ax
	pop dx
	pop cx
	pop ds			; restore all
	lleave code
	jmp .chain		; chain to old handler -->

	lleave ctx


	align 16
	endarea handler
